# -*- coding: utf-8 -*-
#
# Copyright © 2012 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public
# License as published by the Free Software Foundation; either version
# 2 of the License (GPLv2) or (at your option) any later version.
# There is NO WARRANTY for this software, express or implied,
# including the implied warranties of MERCHANTABILITY,
# NON-INFRINGEMENT, or FITNESS FOR A PARTICULAR PURPOSE. You should
# have received a copy of GPLv2 along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.

"""
Errata Support for Yum Importer
"""
import os
import time
import yum
from pulp_rpm.common.ids import TYPE_ID_ERRATA, UNIT_KEY_ERRATA, METADATA_ERRATA, UNIT_KEY_RPM
from pulp_rpm.yum_plugin import util, updateinfo
from pulp.server.db.model.criteria import UnitAssociationCriteria, Criteria
from yum_importer import importer_rpm

_LOG = util.getLogger(__name__)

def get_available_errata(repo_dir):
    """
        Check and Parses the updateinfo.xml and extract errata items to sync

        @param repo_dir repo directory where the metadata can be found
        @type repo_dir str

        @return a dict of errata items with errata id as key
        @rtype {'':{}}
    """
    repomd_xml = os.path.join(repo_dir, "repodata/repomd.xml")
    ftypes = util.get_repomd_filetypes(repomd_xml)
    errata_from_xml = {}
    if "updateinfo" not in ftypes:
        _LOG.info("Unable to find 'updateinfo' in %s" % (ftypes))
        return errata_from_xml
    updateinfo_xml_path = os.path.join(repo_dir, util.get_repomd_filetype_path(repomd_xml, "updateinfo"))
    if not os.path.exists(updateinfo_xml_path):
        _LOG.info("Unable to find errata file at %s")
        return {}
    try:
        errata_from_xml = updateinfo.get_errata(updateinfo_xml_path)
    except yum.Errors.YumBaseError, e:
        _LOG.error("Error parsing updateinfo file [%s]; Error: %s" % (updateinfo_xml_path, e))
    errata_items = {}
    for e_obj in errata_from_xml:
        errata_items[e_obj['id']] = e_obj
    return errata_items

def get_existing_errata(sync_conduit, criteria=None):
    """
     Lookup existing erratum type units in pulp

     @param sync_conduit
     @type sync_conduit pulp.server.content.conduits.repo_sync.RepoSyncConduit

     @param criteria
     @type criteria pulp.server.managers.repo.unit_association_query.UnitAssociationCriteria

     @return a dictionary of existing units, key is the errata id and the value is the unit
     @rtype {():pulp.server.content.plugins.model.Unit}
    """
    existing_units = {}
    for u in sync_conduit.get_units(criteria):
        key = u.unit_key['id']
        existing_units[key] = u
    return existing_units

def get_orphaned_errata(available_errata, existing_errata):
    """
    @param available_errata a dict of errata
    @type available_errata {}

    @param existing_errata dict of units
    @type existing_errata {key:pulp.server.content.plugins.model.Unit}

    @return a dictionary of orphaned units, key is the errata id and the value is the unit
    @rtype {key:pulp.server.content.plugins.model.Unit}
    """
    orphaned_errata = {}
    for key in existing_errata:
        if key not in available_errata:
            orphaned_errata[key] = existing_errata[key]
    return orphaned_errata


def get_new_errata_units(available_errata, sync_conduit):
    """
    Determines which errata to add  or remove and will initialize new units

    :param available_errata: a dict of available errata
    :type  available_errata: dict
    :param sync_conduit:     The sync conduit to save new units to
    :type  sync_conduit:     pulp.server.content.conduits.repo_sync.RepoSyncConduit
    :return:                  A tuple of 2 dictionaries and the sync conduit.  First dict is of new
                             errata, second dict is of new units, and I (rbarlow) am not sure why we
                             are returning the sync conduit, but apparently we are.
    :rtype:                  tuple
    """
    new_errata = {}
    new_units = {}
    for key in available_errata:
        criteria = Criteria(filters={'id' : key})
        try:
            # errata id is unique and will return only 1 entry if matched
            existing_erratum = sync_conduit.search_all_units(type_id=TYPE_ID_ERRATA,
                                                             criteria=criteria)[0]
        except IndexError:
            existing_erratum = None
        if existing_erratum:
            if available_errata[key]['updated'] and existing_erratum.metadata['updated']:
                available_errata_date = available_errata[key]['updated']
                existing_errata_date = existing_erratum.metadata['updated']
            else:
                # updated date is missing, lets use issued date instead
                available_errata_date = available_errata[key]['issued']
                existing_errata_date = existing_erratum.metadata['issued']

            if available_errata_date < existing_errata_date:
                # erratum we have is already newer, skip to the next one
                continue
            elif available_errata_date == existing_errata_date:
                # Its the same errata as we already have, but the pkglist collections could be
                # different. compare the collection name in the list of what we already have
                # if the collection name is missing we add it to delta.
                coll_names = [plist['name'] for plist in existing_erratum.metadata['pkglist']]
                for elist in available_errata[key]['pkglist']:
                    if elist['name'] not in coll_names:
                        # merge the pkglist and save
                        existing_erratum.metadata['pkglist'].append(elist)
                        new_units[key] = existing_erratum
                        new_errata[key] = available_errata[key]
                # We need to save this erratum, so that the repo_content_units collection is sure to
                # have an entry for the repo for this sync_conduit
                sync_conduit.save_unit(existing_erratum)
                continue
        # If we're here, the existing erratum is outdated or doesnt exist. Let's create/update
        # new available erratum.
        erratum = available_errata[key]
        new_errata[key] = erratum
        unit_key  = form_errata_unit_key(erratum)
        metadata =  form_errata_metadata(erratum)
        new_units[key] = sync_conduit.init_unit(TYPE_ID_ERRATA, unit_key, metadata, None)
    return new_errata, new_units, sync_conduit


def form_errata_unit_key(erratum):
    unit_key = {}
    for key in UNIT_KEY_ERRATA:
        unit_key[key] = erratum[key]
    return unit_key

def form_errata_metadata(erratum):
    metadata = {}
    for key in METADATA_ERRATA:
        metadata[key] = erratum[key]
    return metadata

def errata_sync_details(errata_list):
    errata_details = dict()
    errata_details['types'] = {'bugfix' : [], 'security' : [], 'enhancement' : [], }
    for erratum in errata_list.values():
        if erratum['type'] == 'bugfix':
            errata_details['types']["bugfix"].append(erratum)
        elif erratum['type'] == 'security':
            errata_details['types']["security"].append(erratum)
        elif erratum['type'] == 'enhancement':
            errata_details['types']["enhancement"].append(erratum)
    return errata_details

def link_errata_rpm_units(sync_conduit, new_errata_units):
    """
    Links errata to corresponding rpms

    @param sync_conduit
    @type sync_conduit pulp.server.content.conduits.repo_sync.RepoSyncConduit

    @param new_errata_units: errata units to link
    @type  new_errata_units: {}

    @return a link_report dictionry.
    @rtype {}
    """
    valid_fields = []
    valid_fields.extend(UNIT_KEY_RPM)
    valid_fields.append("_storage_path")
    link_report = {}
    existing_rpms = {}
    # link errata and rpm units
    criteria = UnitAssociationCriteria(type_ids=[importer_rpm.RPM_TYPE_ID], unit_fields=valid_fields)
    partial_rpms = importer_rpm.get_existing_units(sync_conduit, criteria=criteria)
    existing_rpms.update(partial_rpms)
    criteria = UnitAssociationCriteria(type_ids=[importer_rpm.SRPM_TYPE_ID], unit_fields=valid_fields)
    partial_rpms = importer_rpm.get_existing_units(sync_conduit, criteria=criteria)
    existing_rpms.update(partial_rpms)
    link_report['linked_units'] = []
    link_report['missing_rpms'] = []
    for u in new_errata_units.values():
        pkglist = u.metadata['pkglist']
        for pkg in pkglist:
            for pinfo in pkg['packages']:
                if not pinfo.has_key('sum'):
                    _LOG.debug("Missing checksum info on package <%s> for linking a rpm to an erratum." % (pinfo))
                    continue
                pinfo['checksumtype'], pinfo['checksum'] = pinfo['sum']
                rpm_key = importer_rpm.form_lookup_key(pinfo)
                if rpm_key in existing_rpms.keys():
                    rpm_unit = existing_rpms[rpm_key]
                    sync_conduit.link_unit(u, rpm_unit, bidirectional=True)
                    link_report['linked_units'].append(rpm_unit)
                else:
                    link_report['missing_rpms'].append(pinfo)
                    _LOG.debug("rpm unit %s not found; skipping" % pinfo)
    return link_report

class ErrataProgress(object):
    def __init__(self, step="Importing Errata"):
        self.step = step
        self.details = {}

class ImporterErrata(object):
    def __init__(self):
        self.canceled = False

    def cancel_sync(self):
        self.canceled = True

    def sync(self, repo, sync_conduit, config, importer_progress_callback=None):
        """
          Invokes errata sync sequence

          @param repo: metadata describing the repository
          @type  repo: L{pulp.server.content.plugins.data.Repository}

          @param sync_conduit
          @type sync_conduit pulp.server.content.conduits.repo_sync.RepoSyncConduit

          @param config: plugin configuration
          @type  config: L{pulp.server.content.plugins.config.PluginCallConfiguration}

          @param importer_progress_callback callback to report progress info to sync_conduit
          @type importer_progress_callback function

          @return a tuple of state, dict of sync summary and dict of sync details
          @rtype (bool, {}, {})
        """
        def set_progress(status):
            if importer_progress_callback:
                importer_progress_callback("errata", status)

        if self.canceled:
            return False, {}, {}
        skip_list = config.get("skip") or []
        if 'erratum' in skip_list:
            _LOG.info("errata content type part of skip types; skipping import")
            return True, {}, {}

        start = time.time()
        repo_dir = "%s/%s" % (repo.working_dir, repo.id)
        available_errata = get_available_errata(repo_dir)
        _LOG.info("Available Errata %s" % len(available_errata))
        progress = {"state":"IN_PROGRESS", "num_errata":len(available_errata)}
        set_progress(progress)

        criteria = UnitAssociationCriteria(type_ids=[TYPE_ID_ERRATA])
        existing_errata = get_existing_errata(sync_conduit, criteria=criteria)
        orphaned_units = get_orphaned_errata(available_errata, existing_errata)
        new_errata, new_units, sync_conduit = get_new_errata_units(available_errata, sync_conduit)
        _LOG.info("%s new_errata, %s new_units" % (len(new_errata), len(new_units)))
        # Save the new units
        for u in new_units.values():
            sync_conduit.save_unit(u)

        # clean up any orphaned errata
        for u in orphaned_units.values():
            _LOG.debug("Removing orphaned unit: %s " % (u))
            sync_conduit.remove_unit(u)
        # link errata with rpm units
        link_report = link_errata_rpm_units(sync_conduit, new_units)

        # get errata sync details
        errata_details = errata_sync_details(new_errata)
        end = time.time()

        summary = dict()
        summary["num_new_errata"] = len(new_errata)
        summary["num_existing_errata"] = len(existing_errata)
        summary["num_orphaned_errata"] = len(orphaned_units)
        summary["errata_time_total_sec"] = end - start

        details = dict()
        details["num_bugfix_errata"] = len(errata_details['types']['bugfix'])
        details["num_security_errata"] = len(errata_details['types']['security'])
        details["num_enhancement_errata"] = len(errata_details['types']['enhancement'])
        details['link_report'] = link_report
        _LOG.debug("Errata Summary: %s \n Details: %s" % (summary, details))
        progress = {"state":"FINISHED", "num_errata":len(available_errata)}
        set_progress(progress)
        _LOG.info("Finished errata sync")
        return True, summary, details
